@@ -34,6 +34,8 @@
 #include <gdk/gdkx.h>
 #include <X11/Xatom.h>
 
+#include "private.h"
+
 #define DISPLAY(o) ((o)->info->screen->xdisplay)
 
 typedef struct ScreenInfo ScreenInfo;
@@ -362,11 +364,7 @@ fill_out_screen_info (Display *xdisplay,
 	     screen->min_height, screen->max_height);
 #endif
     
-#if (RANDR_MAJOR == 1 && RANDR_MINOR >= 3)
-    resources = XRRGetScreenResourcesCurrent (xdisplay, xroot);
-#else
     resources = XRRGetScreenResources (xdisplay, xroot);
-#endif
     
     if (resources)
     {
@@ -541,6 +539,8 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 
     g_return_val_if_fail (error == NULL || *error == NULL, NULL);
     
+    _gnome_desktop_init_i18n ();
+
     if (XRRQueryExtension (dpy, &event_base, &ignore))
     {
 	GnomeRRScreen *screen = g_new0 (GnomeRRScreen, 1);
